{$I-,S-}
UNIT GCrt;

INTERFACE

USES GUtils;

CONST
  BW40     = $00;  CO40     = $01;
  BW80     = $02;  CO80     = $03;
  C40      = CO40; C80      = CO80;
  Mono     = $07;  Font8x8  = $7F;

  Black    = $00;  DarkGray    = $08;
  Blue     = $01;  LightBlue   = $09;
  Green    = $02;  LightGreen  = $0A;
  Cyan     = $03;  LightCyan   = $0B;
  Red      = $04;  LightRed    = $0C;
  Magenta  = $05;  LightMagenta= $0D;
  Brown    = $06;  Yellow      = $0E;
  LightGray= $07;  White       = $0F;

  Blink    = $80;

TYPE
  Wind = RECORD
    X,Y: Byte;
  END;

VAR
  LastMode: Byte;
  TextAttr: Byte;
  WindMin : Wind;
  WindMax : Wind;
  WhereX  : Byte;
  WhereY  : Byte;

FUNCTION  KeyPressed    : Boolean;
FUNCTION  ReadKey       : Char;
PROCEDURE TextMode      (Mode:Byte);
PROCEDURE Window        (X1,Y1,X2,Y2:Byte);
PROCEDURE GotoXY        (X,Y:Byte);
PROCEDURE ClrScr;
PROCEDURE ClrEol;
PROCEDURE InsLine;
PROCEDURE DelLine;
PROCEDURE FWrite        (St:STRING);
PROCEDURE FWriteLn      (St:STRING);

PROCEDURE TextColor     (Color:Byte);
PROCEDURE TextBackground(Color:Byte);
PROCEDURE LowVideo;     INLINE($80/$26/TextAttr/$F7);
PROCEDURE HighVideo;    INLINE($80/$0E/TextAttr/$08);
PROCEDURE NormVideo;    INLINE($B0/$07/$A2/TextAttr);
PROCEDURE Delay         (MS:Word);
PROCEDURE Sound         (Hz:Word);
PROCEDURE NoSound;



IMPLEMENTATION


CONST
  Up   = 1;
  Down = 2;

VAR
  ScanCode : Byte;

 PROCEDURE MoveScreen(X1,Y1,X2,Y2,Lines,Direction:Byte);
 VAR Size:Word; ScreenPtr:Pointer; St:STRING;
 BEGIN
   St := Replicate(Succ(X2-X1),' ');
   Size := Succ(Y2-Y1)*Succ(X2-X1)*2;
   GetMem(ScreenPtr,Size);
   PartSave(X1,Y1,X2,Y2,ScreenPtr^);
   CASE Direction OF
      Up: BEGIN
            PartRestore(X1,Y1-Lines,X2,Pred(Y2+Lines),ScreenPtr^);
            ClearText(X1,Y2,X2,Pred(Y2+Lines),TextAttr);
          END;
    Down: BEGIN
            PartRestore(X1,Y1+Lines,X2,Y2+Lines,ScreenPtr^);
            ClearText(X1,Y1,X2,Pred(Y1+Lines),TextAttr);
          END;
   END;
   FreeMem(ScreenPtr,Size);
 END;

 PROCEDURE ScreenUP;
 BEGIN
   MoveScreen(Succ(WindMin.X),Succ(WindMin.Y)+1,
              Succ(WindMax.X),Succ(WindMax.Y),1,Up);
 END;




{ ------------------------------------------------------------------------ }
{ ------------------------------------------------------------------------ }
{ ------------------------------------------------------------------------ }
 FUNCTION KeyPressed: Boolean; ASSEMBLER;
 ASM
	CMP	ScanCode,0
	JNE	@@1
	MOV	AH,1
	INT	16H
	MOV	AL,0
	JE	@@2
 @@1:	MOV	AL,1
 @@2:
 END;

 FUNCTION ReadKey: Char; ASSEMBLER;
 ASM
	MOV	AL,ScanCode
	MOV	ScanCode,0
	OR	AL,AL
	JNE	@@1
	XOR	AH,AH
	INT	16H
	OR	AL,AL
	JNE	@@1
	MOV	ScanCode,AH
 	OR	AH,AH
 @@1:
 END;

 PROCEDURE TextMode(Mode:Byte);
 BEGIN
   IF Mode >= $7F
     THEN SetCondensedLines
     ELSE ASM
		MOV	AH, 0
		MOV	AL, Mode
		INT	10h
          END;
   TextAttr  := 7;
   WhereX    := 1;
   WhereY    := 1;
   WindMin.X := 0;
   WindMin.Y := 0;
   WindMax.X := MaxX-1;
   WindMax.Y := MaxY-1;
   LastMode  := Mode;
 END;

 PROCEDURE Window(X1,Y1,X2,Y2:Byte);
 BEGIN
   WindMin.X := Pred(X1);
   WindMin.Y := Pred(Y1);
   WindMax.X := Pred(X2);
   WindMax.Y := Pred(Y2);
   GotoXY(1,1);
 END;

 PROCEDURE GotoXY(X,Y: Byte); ASSEMBLER;
 ASM
	MOV	AH, $0F
	INT	10h

	MOV	DL, X
	MOV	DH, Y
	DEC	DL
	ADD	DL, WindMin.X
	JC	@End
	CMP	DL, WindMax.X
	JA	@End

	DEC	DH
	ADD	DH, WindMin.Y
	JC	@End
	CMP	DH, WindMax.Y
	JA	@End

	MOV	AL, X
	MOV	AH, Y
	MOV	WhereX, AL
	MOV	WhereY, AH

	MOV	AH, $02
	INT	10h
 @End:
 END;

 PROCEDURE ClrScr;
 BEGIN
   ClearText(Succ(WindMin.X),Succ(WindMin.Y),
             Succ(WindMax.X),Succ(WindMax.Y),TextAttr);
   GotoXY(1,1);
 END;

 PROCEDURE ClrEol;
 BEGIN
   FastWrite(WhereX,WhereY,TextAttr,Replicate(Succ(WindMax.X)-WhereX,' '));
 END;

 PROCEDURE DelLine;
 BEGIN
   MoveScreen(Succ(WindMin.X),Succ(WhereY+WindMin.Y),
              Succ(WindMax.X),Succ(WindMax.Y),1,Up);
 END;

 PROCEDURE InsLine;
 BEGIN
   MoveScreen(Succ(WindMin.X),WhereY+WindMin.Y,
              Succ(WindMax.X),WindMax.Y,1,Down);
 END;

 PROCEDURE FWrite(St:STRING);
 VAR I:Byte;

  PROCEDURE WriteCh(Ch:Char);
  BEGIN
    FastWrite(WhereX+WindMin.X,WhereY+WindMin.Y,TextAttr,Ch);
  END;

 BEGIN
   FOR I := 1 TO Length(St) DO
   BEGIN
     WriteCh(St[I]);
     IF WhereX+WindMin.X = Succ(WindMax.X)
      THEN BEGIN
             WhereX := 1;
             IF WhereY+WindMin.Y = Succ(WindMax.Y)
               THEN ScreenUp
               ELSE Inc(WhereY);
           END
      ELSE Inc(WhereX);
   END;
   GotoXY(WhereX,WhereY);
 END;

 PROCEDURE FWriteLn(St:STRING);
 BEGIN
   FWrite(St);
   WhereX := 1;
   IF WhereY+WindMin.Y = Succ(WindMax.Y)
     THEN ScreenUp
     ELSE Inc(WhereY);
   GotoXY(WhereX,WhereY);
 END;

 PROCEDURE TextColor(Color:Byte);
 BEGIN
   TextAttr := Attr(Color,BAttr(TextAttr));
 END;

 PROCEDURE TextBackground(Color:Byte);
 BEGIN
   TextAttr := Attr(FAttr(TextAttr),Color);
 END;

 PROCEDURE Delay(MS:Word); ASSEMBLER;
 ASM
	MOV	AX, 1000
	MUL	Ms
	MOV	CX, DX
	MOV	DX, AX
	MOV	AH, $86
	INT	15h
 END;

 PROCEDURE Sound(Hz:Word); ASSEMBLER;
 ASM
	MOV	BX, Hz
	MOV	AX, $34DD
	MOV	DX, $0012
	CMP	DX, BX
	JNC	@End
	DIV	BX
	MOV	BX, AX
	IN	AL, 61h
	TEST	AL, 3
	JNZ	@@1
	OR	AL, 3
	OUT	61H, AL
	MOV	AL, $B6
	OUT	43h, AL
 @@1:	MOV	AL, BL
	OUT	42h, AL
	MOV	AL, BH
	OUT	42h, AL
 @End:
 END;

 PROCEDURE NoSound; ASSEMBLER;
 ASM
	IN	AL, $61
	AND	AL, $FC
	OUT	61h, AL
 END;


BEGIN
  TextAttr  := 7;
  LastMode  := StartMode;
  WhereX    := GUtils.WhereX;
  WhereY    := GUtils.WhereY;
  WindMin.X := 0;
  WindMin.Y := 0;
  WindMax.X := MaxX-1;
  WindMax.Y := MaxY-1;

  ScanCode  := 0;
END.
