{ DEFINE Debug}
{$A+,B-,F+,G+,E-,N-,P+,T-,V-,X+}
{$IFDEF Debug}
  {$D+,L+,Y+,I+,Q-,R+,S+}
{$ELSE}
  {$D-,L-,Y-,I-,Q-,R-,S-}
{$ENDIF}

{{$A+,B-,D+,E-,F+,I-,L+,O-,N-,P+,Q-,R-,S-,T-,V-,X+}

UNIT GMemory;


INTERFACE



CONST
{ Maximum TCollection size }
  MaxCollectionSize = 65520 DIV SizeOf(Pointer); {16380}

{ TCollection error codes }
  coIndexError = -1;              { Index out of range }
  coOverflow   = -2;              { Overflow }

{ VMT header size }
  vmtHeaderSize = 8;

TYPE
{ String pointer }
  PString = ^STRING;

{ TObject base object }
  PObject = ^TObject;
  TObject = OBJECT
    CONSTRUCTOR Init;
    PROCEDURE Free;
    DESTRUCTOR Done; VIRTUAL;
  END;

{ TCollection types }
  PItemList = ^TItemList;
  TItemList = ARRAY[0..MaxCollectionSize - 1] of Pointer;

{ TCollection object }
  PCollection = ^TCollection;
  TCollection = OBJECT(TObject)
    Items: PItemList;
    Count: Integer;
    Limit: Integer;
    Delta: Integer;
    CONSTRUCTOR Init(ALimit, ADelta: Integer);
    DESTRUCTOR Done; VIRTUAL;
    FUNCTION At(Index: Integer): Pointer;
    PROCEDURE AtDelete(Index: Integer);
    PROCEDURE AtFree(Index: Integer);
    PROCEDURE AtInsert(Index: Integer; Item: Pointer);
    PROCEDURE AtPut(Index: Integer; Item: Pointer);
    PROCEDURE Delete(Item: Pointer);
    PROCEDURE DeleteAll;
    PROCEDURE Error(Code, Info: Integer); VIRTUAL;
    FUNCTION FirstThat(Test: Pointer): Pointer;
    PROCEDURE ForEach(Action: Pointer);
    PROCEDURE Free(Item: Pointer);
    PROCEDURE FreeAll;
    PROCEDURE FreeItem(Item: Pointer); VIRTUAL;
    FUNCTION IndexOf(Item: Pointer): Integer; VIRTUAL;
    PROCEDURE Insert(Item: Pointer); VIRTUAL;
    FUNCTION LastThat(Test: Pointer): Pointer;
    PROCEDURE Pack;
    PROCEDURE SetLimit(ALimit: Integer); VIRTUAL;
  END;

{ TSortedCollection OBJECT }
  PSortedCollection = ^TSortedCollection;
  TSortedCollection = OBJECT(TCollection)
    Duplicates: Boolean;
    FUNCTION Compare(Key1,Key2:Pointer): Integer; VIRTUAL;
    FUNCTION IndexOf(Item:Pointer): Integer; VIRTUAL;
    FUNCTION KeyOf(Item:Pointer): Pointer; VIRTUAL;
    FUNCTION Search(Key:Pointer; VAR Index:Integer): Boolean; VIRTUAL;
    PROCEDURE Insert(Item:Pointer); VIRTUAL;
    CONSTRUCTOR Init(ALimit,ADelta:Integer);
  END;

{ TStringCollection OBJECT }
  PStringCollection = ^TStringCollection;
  TStringCollection = OBJECT(TSortedCollection)
    FUNCTION Compare(Key1,Key2:Pointer): Integer; VIRTUAL;
    PROCEDURE FreeItem(Item:Pointer); VIRTUAL;
  END;



{ Dynamic string handling routines }
FUNCTION NewStr(CONST S:STRING): PString;
FUNCTION GetStr(AStr:PString): STRING;
PROCEDURE SetStr(VAR P:PString; S:STRING);
PROCEDURE DisposeStr(P:PString);

FUNCTION Malloc(VAR Ptr; Size:Word): Word;
FUNCTION Dalloc(VAR Ptr): Word;
FUNCTION DosMemAvail: LongInt;





IMPLEMENTATION



PROCEDURE Abstract;
BEGIN
  RunError(211);
END;


{ TObject }
CONSTRUCTOR TObject.Init;
TYPE Image = RECORD
       Link: Word;
       Data: RECORD END;
     END;
BEGIN
  FillChar(Image(Self).Data, SizeOf(Self)-SizeOf(TObject), 0);
END;

{ Shorthand PROCEDURE for a done/dispose }
PROCEDURE TObject.Free;
BEGIN
  Dispose(PObject(@Self), Done);
END;

DESTRUCTOR TObject.Done;
BEGIN
END;




{ TCollection }

CONST
  TCollection_Error    = vmtHeaderSize + $04;
  TCollection_SetLimit = vmtHeaderSize + $14;

PROCEDURE CollectionError; NEAR; ASSEMBLER;
ASM
        PUSH    AX
        PUSH    BX
        PUSH    ES
        PUSH    DI
        MOV     DI,ES:[DI]
        CALL    DWORD PTR [DI].TCollection_Error
END;

CONSTRUCTOR TCollection.Init(ALimit, ADelta: Integer);
BEGIN
  TObject.Init;
  Items := NIL;
  Count := 0;
  Limit := 0;
  Delta := ADelta;
  SetLimit(ALimit);
END;

DESTRUCTOR TCollection.Done;
BEGIN
  FreeAll;
  SetLimit(0);
END;

FUNCTION TCollection.At(Index: Integer): Pointer; ASSEMBLER;
ASM
        LES     DI,Self
        MOV     BX,Index
        OR      BX,BX
        JL      @@1
        CMP     BX,ES:[DI].TCollection.Count
        JGE     @@1
        LES     DI,ES:[DI].TCollection.Items
        SHL     BX,1
        SHL     BX,1
        MOV     AX,ES:[DI+BX]
        MOV     DX,ES:[DI+BX+2]
        JMP     @@2
@@1:    MOV     AX,coIndexError
        CALL    CollectionError
        XOR     AX,AX
        MOV     DX,AX
@@2:
END;

PROCEDURE TCollection.AtDelete(Index: Integer); ASSEMBLER;
ASM
        LES     DI,Self
        MOV     BX,Index
        OR      BX,BX
        JL      @@1
        CMP     BX,ES:[DI].TCollection.Count
        JGE     @@1
        DEC     ES:[DI].TCollection.Count
        MOV     CX,ES:[DI].TCollection.Count
        SUB     CX,BX
        JE      @@2
        CLD
        LES     DI,ES:[DI].TCollection.Items
        SHL     BX,1
        SHL     BX,1
        ADD     DI,BX
        LEA     SI,[DI+4]
        SHL     CX,1
        PUSH    DS
        PUSH    ES
        POP     DS
        REP     MOVSW
        POP     DS
        JMP     @@2
@@1:    MOV     AX,coIndexError
        CALL    CollectionError
@@2:
END;

PROCEDURE TCollection.AtFree(Index:Integer);
VAR Item:Pointer;
BEGIN
  Item := At(Index);
  AtDelete(Index);
  FreeItem(Item);
END;

PROCEDURE TCollection.AtInsert(Index: Integer; Item: Pointer); ASSEMBLER;
ASM
        LES     DI,Self
        MOV     BX,Index
        OR      BX,BX
        JL      @@3
        MOV     CX,ES:[DI].TCollection.Count
        CMP     BX,CX
        JG      @@3
        CMP     CX,ES:[DI].TCollection.Limit
        JNE     @@1
        PUSH    CX
        PUSH    BX
        ADD     CX,ES:[DI].TCollection.Delta
        PUSH    CX
        PUSH    ES
        PUSH    DI
        MOV     DI,ES:[DI]
        CALL    DWORD PTR [DI].TCollection_SetLimit
        POP     BX
        POP     CX
        LES     DI,Self
        CMP     CX,ES:[DI].TCollection.Limit
        JE      @@4
@@1:    INC     ES:[DI].TCollection.Count
        STD
        LES     DI,ES:[DI].TCollection.Items
        SHL     CX,1
        ADD     DI,CX
        ADD     DI,CX
        INC     DI
        INC     DI
        SHL     BX,1
        SUB     CX,BX
        JE      @@2
        LEA     SI,[DI-4]
        PUSH    DS
        PUSH    ES
        POP     DS
        REP     MOVSW
        POP     DS
@@2:    MOV     AX,WORD PTR [Item+2]
        STOSW
        MOV     AX,WORD PTR [Item]
        STOSW
        CLD
        JMP     @@6
@@3:    MOV     AX,coIndexError
        JMP     @@5
@@4:    MOV     AX,coOverflow
        MOV     BX,CX
@@5:    CALL    CollectionError
@@6:
END;

PROCEDURE TCollection.AtPut(Index: Integer; Item: Pointer); ASSEMBLER;
ASM
        MOV     AX,Item.Word[0]
        MOV     DX,Item.Word[2]
        LES     DI,Self
        MOV     BX,Index
        OR      BX,BX
        JL      @@1
        CMP     BX,ES:[DI].TCollection.Count
        JGE     @@1
        LES     DI,ES:[DI].TCollection.Items
        SHL     BX,1
        SHL     BX,1
        MOV     ES:[DI+BX],AX
        MOV     ES:[DI+BX+2],DX
        JMP     @@2
@@1:    MOV     AX,coIndexError
        CALL    CollectionError
@@2:
END;

PROCEDURE TCollection.Delete(Item: Pointer);
BEGIN
  AtDelete(IndexOf(Item));
END;

PROCEDURE TCollection.DeleteAll;
BEGIN
  Count := 0;
END;

PROCEDURE TCollection.Error(Code, Info: Integer);
BEGIN
  RunError(212 - Code);
END;

FUNCTION TCollection.FirstThat(Test: Pointer): Pointer; ASSEMBLER;
ASM
        LES     DI,Self
        MOV     CX,ES:[DI].TCollection.Count
        JCXZ    @@2
        LES     DI,ES:[DI].TCollection.Items
@@1:    PUSH    ES
        PUSH    DI
        PUSH    CX
        PUSH    WORD PTR ES:[DI+2]
        PUSH    WORD PTR ES:[DI]
        PUSH    WORD PTR [BP]
        CALL    Test
        POP     CX
        POP     DI
        POP     ES
        OR      AL,AL
        JNE     @@3
        ADD     DI,4
        LOOP    @@1
@@2:    XOR     AX,AX
        MOV     DX,AX
        JMP     @@4
@@3:    MOV     AX,ES:[DI]
        MOV     DX,ES:[DI+2]
@@4:
END;

PROCEDURE TCollection.ForEach(Action: Pointer); ASSEMBLER;
ASM
        LES     DI,Self
        MOV     CX,ES:[DI].TCollection.Count
        JCXZ    @@2
        LES     DI,ES:[DI].TCollection.Items
@@1:    PUSH    ES
        PUSH    DI
        PUSH    CX
        PUSH    WORD PTR ES:[DI+2]
        PUSH    WORD PTR ES:[DI]
        PUSH    WORD PTR [BP]
        CALL    Action
        POP     CX
        POP     DI
        POP     ES
        ADD     DI,4
        LOOP    @@1
@@2:
END;

PROCEDURE TCollection.Free(Item: Pointer);
BEGIN
  Delete(Item);
  FreeItem(Item);
END;

PROCEDURE TCollection.FreeAll;
VAR I:Integer;
BEGIN
  FOR I := 0 TO Count - 1 DO FreeItem(At(I));
  Count := 0;
END;

PROCEDURE TCollection.FreeItem(Item: Pointer);
BEGIN
  IF Item <> NIL THEN Dispose(PObject(Item), Done);
END;

FUNCTION TCollection.IndexOf(Item: Pointer): Integer; ASSEMBLER;
ASM
        MOV     AX,Item.Word[0]
        MOV     DX,Item.Word[2]
        LES     DI,Self
        MOV     CX,ES:[DI].TCollection.Count
        JCXZ    @@3
        LES     DI,ES:[DI].TCollection.Items
        MOV     BX,DI
        SHL     CX,1
        CLD
@@1:    REPNE   SCASW
        JCXZ    @@3
        TEST    CX,1
        JE      @@1
        XCHG    AX,DX
        SCASW
        XCHG    AX,DX
        LOOPNE  @@1
        JNE     @@3
        MOV     AX,DI
        SUB     AX,BX
        SHR     AX,1
        SHR     AX,1
        DEC     AX
        JMP     @@2
@@3:    MOV     AX,-1
@@2:
END;

PROCEDURE TCollection.Insert(Item: Pointer);
BEGIN
  AtInsert(Count, Item);
END;

FUNCTION TCollection.LastThat(Test: Pointer): Pointer; ASSEMBLER;
ASM
        LES     DI,Self
        MOV     CX,ES:[DI].TCollection.Count
        JCXZ    @@2
        LES     DI,ES:[DI].TCollection.Items
        MOV     AX,CX
        SHL     AX,1
        SHL     AX,1
        ADD     DI,AX
@@1:    SUB     DI,4
        PUSH    ES
        PUSH    DI
        PUSH    CX
        PUSH    WORD PTR ES:[DI+2]
        PUSH    WORD PTR ES:[DI]
        PUSH    WORD PTR [BP]
        CALL    Test
        POP     CX
        POP     DI
        POP     ES
        OR      AL,AL
        JNE     @@3
        LOOP    @@1
@@2:    XOR     AX,AX
        MOV     DX,AX
        JMP     @@4
@@3:    MOV     AX,ES:[DI]
        MOV     DX,ES:[DI+2]
@@4:
END;

PROCEDURE TCollection.Pack; ASSEMBLER;
ASM
        LES     DI,Self
        MOV     CX,ES:[DI].TCollection.Count
        JCXZ    @@3
        LES     DI,ES:[DI].TCollection.Items
        MOV     SI,DI
        PUSH    DS
        PUSH    ES
        POP     DS
        CLD
@@1:    LODSW
        XCHG    AX,DX
        LODSW
        MOV     BX,AX
        OR      BX,DX
        JE      @@2
        XCHG    AX,DX
        STOSW
        XCHG    AX,DX
        STOSW
@@2:    LOOP    @@1
        POP     DS
        LES     BX,Self
        SUB     DI,WORD PTR ES:[BX].TCollection.Items
        SHR     DI,1
        SHR     DI,1
        MOV     ES:[BX].TCollection.Count,DI
@@3:
END;

PROCEDURE TCollection.SetLimit(ALimit: Integer);
VAR AItems:PItemList;
BEGIN
  IF ALimit < Count THEN ALimit := Count;
  IF ALimit > MaxCollectionSize then ALimit := MaxCollectionSize;
  IF ALimit <> Limit THEN
  BEGIN
    IF ALimit = 0 THEN AItems := NIL ELSE
    BEGIN
      GetMem(AItems, ALimit * SizeOf(Pointer));
      IF (Count <> 0) AND (Items <> NIL) THEN
        Move(Items^, AItems^, Count * SizeOf(Pointer));
    END;
    IF Limit <> 0 THEN FreeMem(Items, Limit * SizeOf(Pointer));
    Items := AItems;
    Limit := ALimit;
  END;
END;


{ TSortedCollection }
CONSTRUCTOR TSortedCollection.Init(ALimit, ADelta: Integer);
BEGIN
  TCollection.Init(ALimit, ADelta);
  Duplicates := False;
END;

FUNCTION TSortedCollection.Compare(Key1, Key2: Pointer): Integer;
BEGIN
  Abstract;
END;

FUNCTION TSortedCollection.IndexOf(Item: Pointer): Integer;
VAR I:Integer;
BEGIN
  IndexOf := -1;
  IF Search(KeyOf(Item), I) THEN
  BEGIN
    IF Duplicates THEN
      WHILE (I < Count) AND (Item <> Items^[I]) DO Inc(I);
    IF I < Count THEN IndexOf := I;
  END;
END;

PROCEDURE TSortedCollection.Insert(Item: Pointer);
VAR I:Integer;
BEGIN
  IF NOT Search(KeyOf(Item), I) OR Duplicates THEN AtInsert(I, Item);
END;

FUNCTION TSortedCollection.KeyOf(Item: Pointer): Pointer;
BEGIN
  KeyOf := Item;
END;

FUNCTION TSortedCollection.Search(Key:Pointer; VAR Index:Integer): Boolean;
VAR L,H,I,C: Integer;
BEGIN
  Search := False;
  L := 0;
  H := Count - 1;
  WHILE L <= H DO
  BEGIN
    I := (L + H) SHR 1;
    C := Compare(KeyOf(Items^[I]), Key);
    IF C < 0 THEN L := I + 1 ELSE
    BEGIN
      H := I - 1;
      IF C = 0 THEN
      BEGIN
        Search := True;
        IF NOT Duplicates THEN L := I;
      END;
    END;
  END;
  Index := L;
END;


{ TStringCollection }
FUNCTION TStringCollection.Compare(Key1,Key2:Pointer): Integer; ASSEMBLER;
ASM
        PUSH    DS
        CLD
        LDS     SI,Key1
        LES     DI,Key2
        LODSB
        MOV     AH,ES:[DI]
        INC     DI
        MOV     CL,AL
        CMP     CL,AH
        JBE     @@1
        MOV     CL,AH
@@1:    XOR     CH,CH
        REP     CMPSB
        JE      @@2
        MOV     AL,DS:[SI-1]
        MOV     AH,ES:[DI-1]
@@2:    SUB     AL,AH
        SBB     AH,AH
        POP     DS
END;

PROCEDURE TStringCollection.FreeItem(Item:Pointer);
BEGIN
  DisposeStr(Item);
END;


{ PString routines }
FUNCTION NewStr(CONST S:STRING): PString;
VAR P:PString;
BEGIN
  IF S = ''
    THEN P := NIL
    ELSE BEGIN
           GetMem(P, Length(S) + 1);
           P^ := S;
         END;
  NewStr := P;
END;


FUNCTION GetStr(AStr:PString): STRING;
BEGIN
  IF AStr = NIL THEN GetStr := '' ELSE GetStr := AStr^;
END;

PROCEDURE SetStr(VAR P:PString; S:STRING);
BEGIN
  IF P = NIL
    THEN P := NewStr(S)
    ELSE BEGIN
           DisposeStr(P);
           P := NewStr(S);
         END;
END;

PROCEDURE DisposeStr(P:PString);
BEGIN
  IF P <> NIL THEN FreeMem(P, Length(P^)+1);
END;




{************************************************************************}
{************************************************************************}
FUNCTION Malloc(VAR Ptr; Size:Word): Word;
BEGIN
   INLINE(
     $8B/$46/<SIZE/         {            mov         ax,[bp+<Size]}
     $B9/$04/$00/           {            mov         cx,4}
     $D3/$E8/               {            shr         ax,cl}
     $40/                   {            inc         ax}
     $89/$C3/               {            mov         bx,ax}
     $B4/$48/               {            mov         ah,$48}
     $CD/$21/               {            int         $21             ;Allocate memory}
     $72/$07/               {            jc          AllocErr        ;If any errors ....}
     $C7/$46/$FE/$00/$00/   {NoErrors:   mov word    [bp-2],0        ;Return 0 for successful allocation}
     $EB/$05/               {            jmp short   Exit}
     $89/$46/$FE/           {AllocErr:   mov         [bp-2],ax       ;Return error code}
     $31/$C0/               {            xor         ax,ax           ;Store a NIL value into the ptr}
     $C4/$7E/<PTR/          {Exit:       les         di,[bp+<Ptr]    ;Address of pointer into es:di}
     $50/                   {            push        ax              ;Save the Segment part}
     $31/$C0/               {            xor         ax,ax           ;Offset is always 0}
     $FC/                   {            cld                         ;Make sure direction is upward}
     $AB/                   {            stosw                       ;Store offset of memory block}
     $58/                   {            pop         ax              ;Get back segment part}
     $AB);                  {            stosw                       ;Store segment of memory block}
END;


FUNCTION Dalloc(VAR Ptr): Word;
BEGIN
   IF Pointer(Ptr) <> NIL THEN
   BEGIN
      INLINE(
        $B4/$49/               {            mov         ah,$49}
        $C4/$7E/<PTR/          {            les         di,[bp+<Ptr]}
        $26/$C4/$3D/           {        es: les         di,[di]}
        $CD/$21/               {            int         $21}
        $72/$02/               {            jc          Exit}
        $31/$C0/               {NoError:    xor         ax,ax}
        $89/$46/$FE);          {Exit:       mov         [bp-2],ax}
      Pointer(Ptr) := NIL;
   END {if}
   ELSE
      Dalloc := 0;
END;

FUNCTION DosMemAvail:LongInt;
BEGIN
  INLINE(
    $BB/$FF/$FF/           {         mov         bx,$FFFF}
    $B4/$48/               {         mov         ah,$48}
    $CD/$21/               {         int         $21}
    $89/$D8/               {         mov         ax,bx}
    $B9/$10/$00/           {         mov         cx,16}
    $F7/$E1/               {         mul         cx}
    $89/$46/$FC/           {         mov         [bp-4],ax}
    $89/$56/$FE);          {         mov         [bp-2],dx}
END;


END.
